/**************************************************************************/
/*                           File: dict_align.cc                          */
/**************************************************************************/
/* Project: Proper Noun Pronunciations Dictionary Alignment               */
/* Author: Neeraj Deshmukh                                                */
/**************************************************************************/
 
/*========================================================================
  This file defines the class used in the alignment of the proper nouns
  dictionary
  ========================================================================*/

/*------------------------------------------------------------------------
  include files
  ------------------------------------------------------------------------*/
#include "dict_align.h"

/*------------------------------------------------------------------------
  forward declarations
  ------------------------------------------------------------------------*/
class Back_Pointer;

/*========================================================================
  Back_Pointer
  class definition for the back pointer that keeps track of the best path
  ========================================================================*/
/*----------------------------
  constructors and destructors
  ----------------------------*/
// default constructor
//
Back_Pointer :: Back_Pointer (String phoneme = "", int score = 0, 
			      Back_Pointer *prev = NULL)
{
  // initialize phoneme string
  //
  this_phoneme_d = new char[strlen (phoneme) + 1];
  strcpy (this_phoneme_d, phoneme);
  
  // assign other fields
  //
  best_score_d   = score;
  prev_phoneme_d = prev;

  // update path score
  //
  if (prev_phoneme_d)
    best_score_d += prev_phoneme_d->best_score_d;
}

// copy constructor
//
Back_Pointer :: Back_Pointer (Back_Pointer *new_bkptr)
{
  // copy string
  //
  this_phoneme_d = new char[strlen (new_bkptr->this_phoneme_d) + 1];
  strcpy (this_phoneme_d, new_bkptr->this_phoneme_d);

  // copy other fields
  //
  best_score_d   = new_bkptr->best_score_d;
  prev_phoneme_d = new_bkptr->prev_phoneme_d;
}

// destructor
//
Back_Pointer :: ~Back_Pointer ()
{
  // free memory
  //
  delete this_phoneme_d;
  delete prev_phoneme_d;
}
  
/*----------------------
  input / output methods
  ----------------------*/
// to store a phoneme
//
void Back_Pointer :: put_phone_cc (String phoneme, int score)
{
  // set the phoneme string
  //
  delete this_phoneme_d;
  this_phoneme_d = new char[strlen (phoneme) + 1];
  strcpy (this_phoneme_d, phoneme);

  // modify corresponding score
  //
  best_score_d = score;
}

// get best path score
//
int Back_Pointer :: get_score_cc ()
{
  // output best score
  //
  return (best_score_d);
}

// get the phoneme
//
String Back_Pointer :: get_phone_cc ()
{
  // copy string and return
  //
  String out_phoneme = new char[strlen (this_phoneme_d) + 1];
  strcpy (out_phoneme, this_phoneme_d);
  return (out_phoneme);
}

/*--------------------
  backtrace and output
  --------------------*/
// backtrace and write to file
//
void Back_Pointer :: backtrace_cc (FILE *outf, int length)
{
  // write all the strings in the reverse order by looping 
  // through the back pointers
  //
  String *aligned_phones = new String[length];
  Back_Pointer *temp     = this;
  int pos                = length - 1;
  while (temp != NULL)
    {
      aligned_phones[pos] = new char[strlen (temp->this_phoneme_d) + 1];
      strcpy (aligned_phones[pos], temp->this_phoneme_d);
      pos --;
      temp = temp->prev_phoneme_d;
    }

  // now write them in correct order to output file
  //
  for (int i = 0; i < length; i++)
    {
      fprintf (outf, "%s", aligned_phones[i]);
      if (i < length - 1)
	fprintf (outf, " ");
    }
  fprintf (outf, "\n");
  
  // free memory
  //
  delete [] aligned_phones;
}

/*========================================================================
  definitions of some global functions
  ========================================================================*/
/*--------------------------------------------
  function to open files and test for validity
  --------------------------------------------*/
FILE *open_file_cc (String name, String mode)
{
  FILE *file;
  file = fopen (name, mode);
  if (file == NULL)
    {
      fprintf (stderr,"Cannot open file %s \n", name);
      exit (1);
    }
  return (file);
}

/*----------------------------------------------
  function to write word tooutput file in format
  ----------------------------------------------*/
void write_word_cc (FILE *outf, String word)
{
  int length = strlen (word);
  if (length < 8)
    fprintf (outf, "%s\t\t\t", word);
  else if (length < 16)
    fprintf (outf, "%s\t\t", word);
  else 
    fprintf (outf, "%s\t", word);
}

/*------------------------------------------------------
  function to read the pronunciation from the dictionary
  ------------------------------------------------------*/
String *read_pron_cc (FILE *inf, int *num_phones)
{
  // read in the pronunciations
  //
  String input_pron = new char[50];
  fscanf (inf, "%[\b\t]", input_pron);
  fscanf (inf, "%[^\n]", input_pron);
  
  // now break it into phoneme components and store them
  //
  String *out_pron = new String[20];
  String temp_ph   = strtok (input_pron, " ");
  out_pron[0] = new char[strlen (temp_ph) + 1];
  strcpy (out_pron[0], temp_ph);

  int k  = 0;
  while ((temp_ph = strtok (NULL, " ")))
    {
      k ++;
      out_pron[k] = new char[strlen (temp_ph) + 1];
      strcpy (out_pron[k], temp_ph);
    }
  
  // free memory
  //
  fscanf (inf, "%[\n]", input_pron);
  delete [] input_pron;
  delete temp_ph;

  // return data
  //
  *num_phones = k + 1;
  return (out_pron);  
}

/*-------------------------------------------------------------------
  function to return the score associated with letter - phoneme pairs
  -------------------------------------------------------------------*/
int get_score_cc (char letter = ' ', String phoneme = "")
{
  // score variable
  //
  int score = 0;
  
  // return score if blank phoneme
  //
  if (strcmp (phoneme, "_") == 0)
    {  
      score = 5;
      return (score);
    }
  
  // look at various letter-phone pairs
  //
  switch (letter)
    {
    case 'a':
    case 'A':
      if (strcmp (phoneme, "@") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "&") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "A") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "aU") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "E") == 0)
	{score = 7; break;}
      if (strcmp (phoneme, "ei") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, ">i") == 0)
	{score = 6; break;}
      break;
      
    case 'b':
    case 'B':
      if (strcmp (phoneme, "b") == 0)
	{score = 10; break;}
      break;
      
    case 'c':
    case 'C':
      if (strcmp (phoneme, "tS") == 0)
	{score = 10; break;}
      if (strcmp (phoneme, "S") == 0)
	{score = 10; break;}
      if (strcmp (phoneme, "k") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "s") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "z") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "&k") == 0)
	{score = 6; break;}
      break;

    case 'd':
    case 'D':
      if (strcmp (phoneme, "d") == 0)
	{score = 10; break;}
      if (strcmp (phoneme, "D") == 0)
	{score = 7; break;}
      break;
      
    case 'e':
    case 'E':
      if (strcmp (phoneme, "E") == 0)
	{score = 10; break;}
      if (strcmp (phoneme, "&") == 0)
	{score = 10; break;}
      if (strcmp (phoneme, "ei") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "i:") == 0)
	{score = 7; break;}
      if (strcmp (phoneme, "I") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "u") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "aI") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, ">i") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "j") == 0)
	{score = 6; break;}
      break;
      
    case 'f':
    case 'F':
      if (strcmp (phoneme, "f") == 0)
	{score = 10; break;}
      break;
      
    case 'g':
    case 'G':
      if (strcmp (phoneme, "g") == 0)
	{score = 10; break;}
      if (strcmp (phoneme, "dZ") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "u") == 0)
	{score = 6; break;}
      break;
      
    case 'h':
    case 'H':
      if (strcmp (phoneme, "h") == 0)
	{score = 10; break;}
      if (strcmp (phoneme, "f") == 0)
	{score = 6; break;}
      break;
      
    case 'i':
    case 'I':
      if (strcmp (phoneme, "I") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "i:") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "aI") == 0)
	{score = 7; break;}
      if (strcmp (phoneme, "&") == 0)
	{score = 7; break;}
      if (strcmp (phoneme, "iU") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "ei") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "w") == 0)
	{score = 6; break;}
      break;
      
    case 'j':
    case 'J':
      if (strcmp (phoneme, "dZ") == 0)
	{score = 9; break;}
      if (strcmp (phoneme, "Z") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "j") == 0)
	{score = 7; break;}
      if (strcmp (phoneme, "h") == 0)
	{score = 7; break;}
      if (strcmp (phoneme, "aI") == 0)
	{score = 6; break;}
      break;
      
    case 'k':
    case 'K':
      if (strcmp (phoneme, "k") == 0)
	{score = 7; break;}
      break;
      
    case 'l':
    case 'L':
      if (strcmp (phoneme, "l") == 0)
	{score = 10; break;}
      if (strcmp (phoneme, "j") == 0)
	{score = 6; break;}
      break;
      
    case 'm':
    case 'M':
      if (strcmp (phoneme, "m") == 0)
	{score = 10; break;}
      break;
      
    case 'n':
    case 'N':
      if (strcmp (phoneme, "n") == 0)
	{score = 10; break;}
      if (strcmp (phoneme, "N") == 0)
	{score = 10; break;}
      break;
      
    case 'o':
    case 'O':
      if (strcmp (phoneme, ">") == 0)
	{score = 10; break;}
      if (strcmp (phoneme, "oU") == 0)
	{score = 9; break;}
      if (strcmp (phoneme, ">i") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "A") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "&") == 0)
	{score = 7; break;}
      if (strcmp (phoneme, "aU") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "U") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "u") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "^") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "w") == 0)
	{score = 6; break;}
      break;

    case 'p':
    case 'P':
      if (strcmp (phoneme, "p") == 0)
	{score = 10; break;}
      if (strcmp (phoneme, "v") == 0)
	{score = 6; break;}
      break;

    case 'q':
    case 'Q':
      if (strcmp (phoneme, "k") == 0)
	{score = 10; break;}
      if (strcmp (phoneme, "g") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "tS") == 0)
	{score = 6; break;}
      break;

    case 'r':
    case 'R':
      if (strcmp (phoneme, "3r") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "&r") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "9r") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, ">r") == 0)
	{score = 8; break;}
      break;

    case 's':
    case 'S':
      if (strcmp (phoneme, "s") == 0)
	{score = 9; break;}
      if (strcmp (phoneme, "z") == 0)
	{score = 9; break;}
      if (strcmp (phoneme, "S") == 0)
	{score = 6; break;}
      break;

    case 't':
    case 'T':
      if (strcmp (phoneme, "t") == 0)
	{score = 10; break;}
      if (strcmp (phoneme, "T") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "D") == 0)
	{score = 7; break;}
      if (strcmp (phoneme, "S") == 0)
	{score = 6; break;}
      break;

    case 'u':
    case 'U':
      if (strcmp (phoneme, "^") == 0)
	{score = 9; break;}
      if (strcmp (phoneme, "u") == 0)
	{score = 9; break;}
      if (strcmp (phoneme, "U") == 0)
	{score = 9; break;}
      if (strcmp (phoneme, "w") == 0)
	{score = 7; break;}
      if (strcmp (phoneme, "&") == 0)
	{score = 7; break;}
      if (strcmp (phoneme, "j") == 0)
	{score = 7; break;}
      if (strcmp (phoneme, "iU") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "oU") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "aU") == 0)
	{score = 6; break;}
      break;

    case 'v':
    case 'V':
      if (strcmp (phoneme, "v") == 0)
	{score = 9; break;}
      if (strcmp (phoneme, "w") == 0)
	{score = 6; break;}
      break;

    case 'w':
    case 'W':
      if (strcmp (phoneme, "v") == 0)
	{score = 9; break;}
      if (strcmp (phoneme, "w") == 0)
	{score = 9; break;}
      if (strcmp (phoneme, "u") == 0)
	{score = 6; break;}
      break;

    case 'x':
    case 'X':
      if (strcmp (phoneme, "ks") == 0)
	{score = 9; break;}
      if (strcmp (phoneme, "kS") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "Z") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "tS") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "k") == 0)
	{score = 7; break;}
      if (strcmp (phoneme, "h") == 0)
	{score = 7; break;}
      break;

    case 'y':
    case 'Y':
      if (strcmp (phoneme, "j") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "aI") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, ">i") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "i:") == 0)
	{score = 8; break;}
      if (strcmp (phoneme, "I") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "&") == 0)
	{score = 6; break;}
      break;

    case 'z':
    case 'Z':
      if (strcmp (phoneme, "Z") == 0)
	{score = 9; break;}
      if (strcmp (phoneme, "z") == 0)
	{score = 7; break;}
      if (strcmp (phoneme, "s") == 0)
	{score = 6; break;}
      break;

    case '\'':
      if (strcmp (phoneme, "i:") == 0)
	{score = 6; break;}
      if (strcmp (phoneme, "I") == 0)
	{score = 6; break;}

    case '-':
      if (strcmp (phoneme, "_") == 0)
	{score = 10; break;}
    }

  // return final score
  //
  return (score);
}
